# -*- coding: utf-8 -*-

import datetime
from distutils.dir_util import copy_tree
import hashlib
import json
import pydoc
import logging
import os
import time
import random
from pathlib import Path
from typing import Union, Optional, Dict, Any, Tuple, List, Callable

import pandas as pd
import numpy as np
import tensorflow as tf

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn import metrics
from sklearn.model_selection import cross_validate, TimeSeriesSplit
from sklearn.pipeline import Pipeline

from gordo_components.util import disk_registry
from gordo_components import serializer, __version__, MAJOR_VERSION, MINOR_VERSION
from gordo_components.dataset.dataset import _get_dataset
from gordo_components.dataset.base import GordoBaseDataset
from gordo_components.model.base import GordoBase
from gordo_components.model.utils import metric_wrapper
from gordo_components.workflow.config_elements.normalized_config import NormalizedConfig

logger = logging.getLogger(__name__)


def build_model(
    name: str,
    model_config: dict,
    data_config: Union[GordoBaseDataset, dict],
    metadata: dict,
    evaluation_config: dict = {
        "cv_mode": "full_build",
        "scoring_scaler": "sklearn.preprocessing.RobustScaler",
    },
) -> Tuple[Union[BaseEstimator, None], dict]:
    """
    Build a model and serialize to a directory for later serving.

    Parameters
    ----------
    name: str
        Name of model to be built
    model_config: dict
        Mapping of Model to initialize and any additional kwargs which are to be used in it's initialization.
        Example::

          {'type': 'KerasAutoEncoder',
           'kind': 'feedforward_hourglass'}

    data_config: dict
        Mapping of the Dataset to initialize, following the same logic as model_config.
    metadata: dict
        Mapping of arbitrary metadata data.
    evaluation_config: dict
        Dict of parameters which are exposed to build_model.
            - cv_mode: str
                String which enables three different modes, represented as a key value in evaluation_config:
                * cross_val_only: Only perform cross validation
                * build_only: Skip cross validation and only build the model
                * full_build: Cross validation and full build of the model, default value
            - scoring_scaler: Optional[str]
                Optionally a string which gives the path to a scaler, which will be
                applied on `Y` and `Y-hat` before scoring. It will not be used for
                training or any other part of the model-building, only for the scoring.
                This is usefull if one has multi-dimensional output, where the different
                dimensions have different scale, and one wants to use a scorer which is
                sensitive to this (e.g. MSE).
                If None then no scaling will take place.

                Example::

                    {"cv_mode": "cross_val_only",
                    "scoring_scaler": "sklearn.preprocessing.RobustScaler"}

    Returns
    -------
        Tuple[Optional[sklearn.base.BaseEstimator], dict]
    """
    if evaluation_config.get("seed"):
        seed = evaluation_config["seed"]
        logger.info(f"Setting random seed: '{seed}'")
        tf.random.set_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    # Get the dataset from config
    logger.debug(f"Initializing Dataset with config {data_config}")

    dataset = (
        data_config
        if isinstance(data_config, GordoBaseDataset)
        else _get_dataset(data_config)
    )

    logger.debug("Fetching training data")
    start = time.time()

    X, y = dataset.get_data()

    time_elapsed_data = time.time() - start

    # Get the model and dataset
    logger.debug(f"Initializing Model with config: {model_config}")
    model = serializer.pipeline_from_definition(model_config)

    cv_duration_sec = None

    scores: Dict[str, Any] = dict()
    if evaluation_config["cv_mode"].lower() in ("cross_val_only", "full_build"):

        # Build up a metrics list.
        metrics_list = metrics_from_list(evaluation_config.get("metrics"))

        # Cross validate
        logger.debug("Starting cross validation")
        start = time.time()
        if hasattr(model, "predict"):

            metrics_dict = get_metrics_dict(
                metrics_list, y, scaler=evaluation_config.get("scoring_scaler", None)
            )

            cv = cross_validate(
                model,
                X,
                y,
                scoring=metrics_dict,
                return_estimator=True,
                cv=TimeSeriesSplit(n_splits=3),
            )
            for metric, test_metric in map(lambda k: (k, f"test_{k}"), metrics_dict):
                val = {
                    "fold-mean": cv[test_metric].mean(),
                    "fold-std": cv[test_metric].std(),
                    "fold-max": cv[test_metric].max(),
                    "fold-min": cv[test_metric].min(),
                }
                val.update(
                    {
                        f"fold-{i + 1}": raw_value
                        for i, raw_value in enumerate(cv[test_metric].tolist())
                    }
                )
                scores.update({metric: val})

        else:
            logger.debug("Unable to score model, has no attribute 'predict'.")

        cv_duration_sec = time.time() - start

        # If cross_val_only, return the cv_scores and empty model.
        if evaluation_config["cv_mode"] == "cross_val_only":
            model = None
            metadata["model"] = {
                "cross-validation": {
                    "cv-duration-sec": cv_duration_sec,
                    "scores": scores,
                }
            }
            return model, metadata
    # Train
    logger.debug("Starting to train model.")
    start = time.time()
    model.fit(X, y)
    time_elapsed_model = time.time() - start

    metadata = {"user-defined": metadata}
    metadata["name"] = name
    metadata["dataset"] = dataset.get_metadata()
    utc_dt = datetime.datetime.now(datetime.timezone.utc)
    metadata["model"] = {
        "model-offset": _determine_offset(model, X),
        "model-creation-date": str(utc_dt.astimezone()),
        "model-builder-version": __version__,
        "model-config": model_config,
        "data-query-duration-sec": time_elapsed_data,
        "model-training-duration-sec": time_elapsed_model,
        "cross-validation": {"cv-duration-sec": cv_duration_sec, "scores": scores},
    }

    metadata["model"].update(_get_metadata(model))

    return model, metadata


def get_metrics_dict(
    metrics_list: list,
    y: pd.DataFrame,
    scaler: Optional[Union[TransformerMixin, str]] = None,
) -> dict:

    """
    Given a list of metrics that accept a true_y and pred_y as inputs this returns a
    dictionary with keys in the form '{score}-{tag_name}' for each given target tag
    and '{score}' for the average score across all target tags and folds,
    and values being the callable make_scorer(metric_wrapper(score)). Note: score in
    {score}-{tag_name} is a sklearn's score function name with '_' replaced by '-'
    and tag_name corresponds to given target tag name with ' ' replaced by '-'.

    Parameters
    ----------
    metrics_list: list
        List of sklearn score functions
    y: pd.DataFrame
        Target data
    scaler : Optional[Union[TransformerMixin, str]]
        Scaler which will be fitted on y, and used to transform the data before
        scoring. Useful when the metrics are sensitive to the amplitude of the data, and
        you have multiple targets.


    Returns
    -------
        dict
    """
    if isinstance(scaler, str):
        scaler = serializer.pipeline_from_definition(scaler)

    if scaler:
        logger.debug("Fitting scaler for scoring purpose")
        scaler.fit(y)

    def _score_factory(metric_func=metrics.r2_score, col_index=0):
        def _score_per_tag(y_true, y_pred):
            # This function extracts the score for each given target_tag to
            # use as scoring argument in sklearn cross_validate, as the scoring
            # must return a single value.
            if hasattr(y_true, "values"):
                y_true = y_true.values
            if hasattr(y_pred, "values"):
                y_pred = y_pred.values

            return metric_func(y_true[:, col_index], y_pred[:, col_index])

        return _score_per_tag

    metrics_dict = {}
    for metric in metrics_list:
        for index, col in enumerate(y.columns):
            metric_str = metric.__name__.replace("_", "-")
            metrics_dict.update(
                {
                    metric_str
                    + f'-{col.replace(" ", "-")}': metrics.make_scorer(
                        metric_wrapper(
                            _score_factory(metric_func=metric, col_index=index),
                            scaler=scaler,
                        )
                    )
                }
            )

        metrics_dict.update(
            {metric_str: metrics.make_scorer(metric_wrapper(metric, scaler=scaler))}
        )
    return metrics_dict


def _determine_offset(model: BaseEstimator, X: Union[np.ndarray, pd.DataFrame]) -> int:
    """
    Determine the model's offset. How much does the output of the model differ
    from its input?

    Parameters
    ----------
    model: sklearn.base.BaseEstimator
        Trained model with either ``predict`` or ``transform`` method, preference
        given to ``predict``.
    X: Union[np.ndarray, pd.DataFrame]
        Data to pass to the model's ``predict`` or ``transform`` method.

    Returns
    -------
    int
        The difference between X and the model's output lengths.
    """
    out = model.predict(X) if hasattr(model, "predict") else model.transform(X)
    return len(X) - len(out)


def _save_model_for_workflow(
    model: BaseEstimator, metadata: dict, output_dir: Union[os.PathLike, str]
):
    """
    Save a model according to the expected Argo workflow procedure.

    Parameters
    ----------
    model: BaseEstimator
        The model to save to the directory with gordo serializer.
    metadata: dict
        Various mappings of metadata to save alongside model.
    output_dir: Union[os.PathLike, str]
        The directory where to save the model, will create directories if needed.

    Returns
    -------
    Union[os.PathLike, str]
        Path to the saved model
    """
    os.makedirs(output_dir, exist_ok=True)  # Ok if some dirs exist
    serializer.dump(model, output_dir, metadata=metadata)
    return output_dir


def _get_metadata(model: BaseEstimator, metadata: dict = dict()) -> dict:
    """
    Recursively check for :class:`gordo_components.model.base.GordoBase` in a
    given ``model``. If such a model exists buried inside of a
    :class:`sklearn.pipeline.Pipeline` which is then part of another
    :class:`sklearn.base.BaseEstimator`, this function will return its metadata.

    Parameters
    ----------
    model: BaseEstimator
    metadata: dict
        Any initial starting metadata, but is mainly meant to be used during
        the recursive calls to accumulate any multiple :class:`gordo_components.model.base.GordoBase`
        models found in this model

    Notes
    -----
    If there is a ``GordoBase`` model inside of a ``Pipeline`` which is not the final
    step, this function will not find it.

    Returns
    -------
    dict
        Dictionary representing accumulated calls to :meth:`gordo_components.model.base.GordoBase.get_metadata`
    """
    metadata = metadata.copy()

    # If it's a Pipeline, only need to get the last step, which potentially has metadata
    if isinstance(model, Pipeline):
        final_step = model.steps[-1][1]
        metadata.update(_get_metadata(final_step))
        return metadata

    # GordoBase is simple, having a .get_metadata()
    if isinstance(model, GordoBase):
        metadata.update(model.get_metadata())

    # Continue to look at object values in case, we decided to have a GordoBase
    # which also had a GordoBase as a parameter/attribute, but will satisfy BaseEstimators
    # which can take a GordoBase model as a parameter, which will then have metadata to get
    for val in model.__dict__.values():
        if isinstance(val, Pipeline):
            metadata.update(_get_metadata(val.steps[-1][1]))
        elif isinstance(val, GordoBase) or isinstance(val, BaseEstimator):
            metadata.update(_get_metadata(val))
    return metadata


def calculate_model_key(
    name: str,
    model_config: dict,
    data_config: dict,
    evaluation_config: dict,
    metadata: Optional[dict] = None,
) -> str:
    """
    Calculates a hash-key from a model and data-config.

    Notes
    -----
    Ignores the data_provider key since this is an complicated object.

    Parameters
    ----------
    name: str
        Name of the model
    model_config: dict
        Config for the model. See
        :func:`gordo_components.builder.build_model.build_model`.
    data_config: dict
        Config for the data-configuration. See
        :func:`gordo_components.builder.build_model.build_model`.
    evaluation_config: dict
        Config for the evaluation-configuration. See
        :func:`gordo_components.builder.build_moodel.build_model`.
    metadata: Optional[dict] = None
        Metadata for the models. See
        :func:`gordo_components.builder.build_model.build_model`.

    Returns
    -------
    str:
        A 512 byte hex value as a string based on the content of the parameters.

    Examples
    -------
    >>> len(calculate_model_key(name="My-model", model_config={"model": "something"},
    ... data_config={"tag_list": ["tag1", "tag 2"]}, evaluation_config={"cv_mode": "full_build"} ))
    128
    """
    if metadata is None:
        metadata = {}
    # TODO Fix this when we get a good way of passing data_provider in the yaml/json
    if "data_provider" in data_config:
        logger.warning(
            "data_provider key found in data_config, ignoring it when creating hash"
        )
        data_config = dict(data_config)
        del data_config["data_provider"]

    # Sets a lot of the parameters to json.dumps explicitly to ensure that we get
    # consistent hash-values even if json.dumps changes their default values (and as such might
    # generate different json which again gives different hash)
    json_rep = json.dumps(
        {
            "name": name,
            "model_config": model_config,
            "data_config": data_config,
            "user-defined": metadata,
            "evaluation_config": evaluation_config,
            "gordo-major-version": MAJOR_VERSION,
            "gordo-minor-version": MINOR_VERSION,
        },
        sort_keys=True,
        default=str,
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        cls=None,
        indent=None,
        separators=None,
    )
    logger.debug(f"Calculating model hash key for model: {json_rep}")
    return hashlib.sha3_512(json_rep.encode("ascii")).hexdigest()


def check_cache(model_register_dir: Union[os.PathLike, str], cache_key: str):
    """
    Checks if a model is cached, and returns its path if it exists.

    Parameters
    ----------
    model_register_dir: [os.PathLike, None]
        The register dir where the model lies.
    cache_key: str
        A 512 byte hex value as a string based on the content of the parameters.

     Returns
    -------
    Union[os.PathLike, None]:
        The path to the cached model, or None if it does not exist.
    """
    existing_model_location = disk_registry.get_value(model_register_dir, cache_key)

    # Check that the model is actually there
    if existing_model_location and Path(existing_model_location).exists():
        logger.debug(
            f"Found existing model at path {existing_model_location}, returning it"
        )
        return existing_model_location
    elif existing_model_location:
        logger.warning(
            f"Found that the model-path {existing_model_location} stored in the "
            f"registry did not exist."
        )
        return None
    else:
        logger.info(
            f"Did not find the model with key {cache_key} in the register at "
            f"{model_register_dir}."
        )
        return None


def provide_saved_model(
    name: str,
    model_config: dict,
    data_config: dict,
    metadata: dict,
    output_dir: Union[os.PathLike, str],
    model_register_dir: Union[os.PathLike, str] = None,
    replace_cache=False,
    evaluation_config: dict = {"cv_mode": "full_build"},
) -> Union[os.PathLike, str]:
    """
    Ensures that the desired model exists on disk in `output_dir`, and returns the path
    to it. If `output_dir` exists we assume the model is there (no validation), and
    return that path.


    Builds the model if needed, or finds it among already existing models if
    ``model_register_dir`` is non-None, and we find the model there. If
    `model_register_dir` is set we will also store the model-location of the generated
    model there for future use. Think about it as a cache that is never emptied.

    Parameters
    ----------
    name: str
        Name of the model to be built
    model_config: dict
        Config for the model. See
        :func:`gordo_components.builder.build_model.build_model`.
    data_config: dict
        Config for the data-configuration. See
        :func:`gordo_components.builder.build_model.build_model`.
    metadata: dict
        Extra metadata to be added to the built models if it is built. See
        :func:`gordo_components.builder.build_model.build_model`.
    output_dir: Union[os.PathLike, str]
        A path to where the model will be deposited if it is built.
    model_register_dir:
        A path to a register, see `gordo_components.util.disk_registry`. If this is None
        then always build the model, otherwise try to resolve the model from the
        registry.
    replace_cache: bool
        Forces a rebuild of the model, and replaces the entry in the cache with the new
        model.
    evaluation_config: dict
        Config for the evaluation. See
        :func:`gordo_components.builder.build_model.build_model`.

    Returns
    -------
    Union[os.PathLike, str]:
        Path to the model
    """
    cache_key = calculate_model_key(
        name, model_config, data_config, evaluation_config, metadata=metadata
    )
    if model_register_dir:
        logger.info(
            f"Model caching activated, attempting to read model-location with key "
            f"{cache_key} from register {model_register_dir}"
        )
        if replace_cache:
            logger.info("replace_cache activated, deleting any existing cache entry")
            disk_registry.delete_value(model_register_dir, cache_key)
        else:
            cached_model_location = check_cache(model_register_dir, cache_key)
            if cached_model_location:
                logger.info(
                    f"Found model in cache, copying from {cached_model_location} to "
                    f"new location {output_dir} "
                )
                if cached_model_location == output_dir:
                    return output_dir
                else:
                    try:
                        # Why not shutil.copytree? Because in python <3.7 it causes
                        # errors on Azure NFS, see:
                        # - https://bugs.python.org/issue24564
                        # - https://stackoverflow.com/questions/51616058/shutil-copystat-fails-inside-docker-on-azure/51635427#51635427
                        copy_tree(
                            str(cached_model_location),
                            str(output_dir),
                            preserve_mode=0,
                            preserve_times=0,
                        )
                    except FileExistsError:
                        logger.warning(
                            f"Found that output directory {output_dir} "
                            f"already exists, assuming model is "
                            f"already located there"
                        )
                    return output_dir

    model, metadata = build_model(
        name=name,
        model_config=model_config,
        data_config=data_config,
        metadata=metadata,
        evaluation_config=evaluation_config,
    )
    model_location = _save_model_for_workflow(
        model=model, metadata=metadata, output_dir=output_dir
    )
    logger.info(f"Successfully built model, and deposited at {model_location}")
    if model_register_dir:
        logger.info(f"Writing model-location to model registry")
        disk_registry.write_key(model_register_dir, cache_key, model_location)
    return model_location


def metrics_from_list(metric_list: Optional[List[str]] = None) -> List[Callable]:
    """
    Given a list of metric function paths. ie. sklearn.metrics.r2_score or
    simple function names which are expected to be in the ``sklearn.metrics`` module,
    this will return a list of those loaded functions.

    Parameters
    ----------
    metrics: Optional[List[str]]
        List of function paths to use as metrics for the model
        Defaults to those specified in :class:`gordo_components.workflow.config_components.NormalizedConfig`
        sklearn.metrics.explained_variance_score,
        sklearn.metrics.r2_score,
        sklearn.metrics.mean_squared_error,
        sklearn.metrics.mean_absolute_error

    Returns
    -------
    List[Callable]
        A list of the functions loaded

    Raises
    ------
    AttributeError:
       If the function cannot be loaded.
    """
    defaults = NormalizedConfig.DEFAULT_CONFIG_GLOBALS["evaluation"]["metrics"]
    funcs = list()
    for func_path in metric_list or defaults:
        func = pydoc.locate(func_path)
        if func is None:
            # Final attempt, load function from sklearn.metrics module.
            funcs.append(getattr(metrics, func_path))
        else:
            funcs.append(func)
    return funcs



# -*- coding: utf-8 -*-
import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Iterable, List, Optional
from urllib.parse import quote

import numpy as np
import pandas as pd
from azure.datalake.store import core

from gordo_components.data_provider.base import GordoBaseDataProvider
from gordo_components.dataset.sensor_tag import SensorTag

logger = logging.getLogger(__name__)


class NcsReader(GordoBaseDataProvider):
    ASSET_TO_PATH = {
        "1101-sfb": "/raw/corporate/IMS Statfjord/sensordata/1101-SFB",
        "1110-gfa": "/raw/corporate/Aspen MS - IP21 Gullfaks A/sensordata/1110-GFA",
        "1125-kvb": "/raw/corporate/PI System Operation Norway/sensordata/1125-KVB",
        "1130-troa": "/raw/corporate/Aspen MS - IP21 Troll A/sensordata/1130-TROA",
        "1138-val": "/raw/corporate/PI System Operation Norway/sensordata/1138-VAL",
        "1163-gdr": "/raw/corporate/PI System Manager Sleipner/sensordata/1163-GDR",
        "1180-nor": "/raw/corporate/PI System Operation North/sensordata/1180-NOR",
        "1190-asga": "/raw/corporate/PI System Operation North/sensordata/1190-ASGA",
        "1191-asgb": "/raw/corporate/PI System Operation North/sensordata/1191-ASGB",
        "1218-gkr": "/raw/corporate/PI System Manager Sleipner/sensordata/1218-GKR",
        "1230-vis": "/raw/corporate/Aspen MS - IP21 Visund/sensordata/1230-VIS",
        "1295-pera": "/raw/corporate/Aspen MS - IP21 Peregrino/sensordata/1295-PERA",
        "1755-gra": "/raw/corporate/Aspen MS - IP21 Grane/sensordata/1755-GRA",
        "1775-trob": "/raw/corporate/Aspen MS - IP21 Troll B/sensordata/1775-TROB",
        "1776-troc": "/raw/corporate/Aspen MS - IP21 Troll C/sensordata/1776-TROC",
        "1900-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1900-JSV",
        "1901-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1901-JSV",
        "1902-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1902-JSV",
        "1903-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1903-JSV",
    }

    def __init__(
        self,
        client: core.AzureDLFileSystem,
        threads: Optional[int] = 1,
        remove_status_codes: Optional[list] = [0],
        dl_base_path: Optional[str] = None,
        **kwargs,
    ):
        """
        Creates a reader for tags from the Norwegian Continental Shelf. Currently
        only supports a small subset of assets.

        Parameters
        ----------
        threads : Optional[int]
            Number of threads to use. If None then use 1 thread
        remove_status_codes: Optional[list]
            Removes data with Status code(s) in the list. By default it removes data
            with Status code 0.
        dl_base_path: Optional[str]
            Base bath used to override the asset to path dictionary. Useful for demos
            and other non-production settings.

        """
        super().__init__(**kwargs)
        self.client = client
        self.threads = threads
        self.remove_status_codes = remove_status_codes
        self.dl_base_path = dl_base_path
        logger.info(f"Starting NCS reader with {self.threads} threads")

    def can_handle_tag(self, tag: SensorTag):
        """
        Implements GordoBaseDataProvider, see base class for documentation
        """
        return (
            self.dl_base_path is not None
            or NcsReader.base_path_from_asset(tag.asset) is not None
        )

    def load_series(
        self,
        from_ts: datetime,
        to_ts: datetime,
        tag_list: List[SensorTag],
        dry_run: Optional[bool] = False,
    ) -> Iterable[pd.Series]:
        """
        See GordoBaseDataProvider for documentation
        """
        if to_ts < from_ts:
            raise ValueError(
                f"NCS reader called with to_ts: {to_ts} before from_ts: {from_ts}"
            )
        adls_file_system_client = self.client

        years = range(from_ts.year, to_ts.year + 1)

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            fetched_tags = executor.map(
                lambda tag: self.read_tag_files(
                    adls_file_system_client=adls_file_system_client,
                    tag=tag,
                    years=years,
                    dry_run=dry_run,
                    remove_status_codes=self.remove_status_codes,
                    dl_base_path=self.dl_base_path,
                ),
                tag_list,
            )

            for tag_frame_all_years in fetched_tags:
                filtered = tag_frame_all_years[
                    (tag_frame_all_years.index >= from_ts)
                    & (tag_frame_all_years.index < to_ts)
                ]
                yield filtered

    @staticmethod
    def read_tag_files(
        adls_file_system_client: core.AzureDLFileSystem,
        tag: SensorTag,
        years: range,
        dry_run: Optional[bool] = False,
        remove_status_codes: Optional[list] = [0],
        dl_base_path: Optional[str] = None,
    ) -> pd.Series:
        """
        Download tag files for the given years into dataframes,
        and return as one dataframe.

        Parameters
        ----------
        adls_file_system_client: core.AzureDLFileSystem
            the AzureDLFileSystem client to use
        tag: SensorTag
            the tag to download data for
        years: range
            range object providing years to include
        dry_run: Optional[bool]
            if True, don't download data, just check info, log, and return
        remove_status_codes: Optional[list]
            Removes data with Status code(s) in the list. By default it removes data
            with Status code 0.
        dl_base_path: Optional[str]
            Base bath used to override the asset to path dictionary. Useful for demos
            and other non-production settings.

        Returns
        -------
        pd.Series:
            Series with all years for one tag.
        """
        tag_base_path = (
            dl_base_path if dl_base_path else NcsReader.base_path_from_asset(tag.asset)
        )

        if not tag_base_path:
            raise ValueError(f"Unable to find base path from tag {tag} ")
        all_years = []
        logger.info(f"Downloading tag: {tag} for years: {years}")

        for year in years:
            tag_name_encoded = quote(tag.name, safe=" ")
            file_path = (
                f"{tag_base_path}/{tag_name_encoded}/{tag_name_encoded}_{year}.csv"
            )
            logger.info(f"Parsing file {file_path}")

            info = adls_file_system_client.info(file_path)
            file_size = info.get("length") / (1024 ** 2)
            logger.info(f"File size: {file_size:.2f}MB")
            if dry_run:
                logger.info("Dry run only, returning empty frame early")
                return pd.DataFrame()

            with adls_file_system_client.open(file_path, "rb") as f:
                df = pd.read_csv(
                    f,
                    sep=";",
                    header=None,
                    names=["Sensor", tag.name, "Timestamp", "Status"],
                    usecols=[tag.name, "Timestamp", "Status"],
                    dtype={tag.name: np.float32},
                    parse_dates=["Timestamp"],
                    date_parser=lambda col: pd.to_datetime(col, utc=True),
                    index_col="Timestamp",
                )

                df = df[~df["Status"].isin(remove_status_codes)]
                all_years.append(df)
                logger.info(f"Done parsing file {file_path}")

        combined = pd.concat(all_years)

        # There often comes duplicated timestamps, keep the last
        if combined.index.duplicated().any():
            combined = combined[~combined.index.duplicated(keep="last")]

        return combined[tag.name]

    @staticmethod
    def base_path_from_asset(asset: str):
        """
        Resolves an asset code to the datalake basepath containing the data.
        Returns None if it does not match any of the asset codes we know.
        """
        if not asset:
            return None

        logger.info(f"Looking for match for asset {asset}")
        asset = asset.lower()
        if asset not in NcsReader.ASSET_TO_PATH:
            logger.info(
                f"Could not find match for asset {asset} in the list of "
                f"supported assets: {NcsReader.ASSET_TO_PATH.keys()}"
            )
            return None

        logger.info(
            f"Found asset code {asset}, returning {NcsReader.ASSET_TO_PATH[asset]}"
        )
        return NcsReader.ASSET_TO_PATH[asset]



# -*- coding: utf-8 -*-

import abc
import itertools
import logging
import time
import typing
from typing import Optional

import pandas as pd

from gordo_components.client.utils import influx_client_from_uri, EndpointMetadata


"""
Module contains objects which can be made into generators which take
and EndpointMetadata and metadata (dict) when instantiated and are sent
prediction dataframes as the prediction client runs::

    def my_forwarder(
        predictions: pd.DataFrame = None,
        endpoint: EndpointMetadata = None,
        metadata: dict = dict(),
        resampled_sensor_data: pd.DataFrame = None
    ):
        ...

The gordo_components.client.utils.EndpointMetadata hold information about 
what endpoint the coroutine should concern itself with, and metadata are 
keyvalue pairs
"""


logger = logging.getLogger(__name__)


class PredictionForwarder(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __call__(
        self,
        *,
        predictions: pd.DataFrame = None,
        endpoint: EndpointMetadata = None,
        metadata: dict = dict(),
        resampled_sensor_data: pd.DataFrame = None,
    ):
        ...


class ForwardPredictionsIntoInflux(PredictionForwarder):
    """
    To be used as a 'forwarder' for the prediction client

    After instantiation, it is a coroutine which accepts prediction dataframes
    which it will pass onto influx
    """

    def __init__(
        self,
        destination_influx_uri: Optional[str] = None,
        destination_influx_api_key: Optional[str] = None,
        destination_influx_recreate: bool = False,
        n_retries=5,
    ):
        """
        Create an instance which, when called, is a coroutine capable of
        being sent dataframes generated from the '/anomaly/prediction' endpoint

        Parameters
        ----------
        destination_influx_uri: str
            Connection string for destination influx -
            format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>
        destination_influx_api_key: str
            API key if needed for destination db
        destination_influx_recreate: bool
            Drop the database before filling it with data?
        """
        # Create df client if provided
        self.n_retries = n_retries
        self.dataframe_client = (
            influx_client_from_uri(
                destination_influx_uri,
                api_key=destination_influx_api_key,
                recreate=destination_influx_recreate,
                dataframe_client=True,
            )
            if destination_influx_uri
            else None
        )

    def __call__(
        self,
        *,
        predictions: pd.DataFrame = None,
        endpoint: EndpointMetadata = None,
        metadata: dict = dict(),
        resampled_sensor_data: pd.DataFrame = None,
    ):
        if resampled_sensor_data is None and predictions is None:
            raise ValueError(
                "Argument `resampled_sensor_data` or `predictions` must be passed"
            )
        if predictions is not None:
            if endpoint is None:
                raise ValueError(
                    "Argument `endpoint`must be provided if `predictions` is provided"
                )
            self.forward_predictions(predictions, endpoint=endpoint, metadata=metadata)
        if resampled_sensor_data is not None:
            self.send_sensor_data(resampled_sensor_data)

    def forward_predictions(
        self,
        predictions: pd.DataFrame,
        endpoint: EndpointMetadata,
        metadata: dict = dict(),
    ):
        """
        Takes a multi-layed column dataframe and write points to influx where
        each top level name is treated as the measurement name.

        Parameters
        ----------
        predictions: pd.DataFrame
            Multi layed column dataframe, where top level names will be treated
            as the 'measurement' name in influx and 2nd level will be the fields
            under those measurements.

        Returns
        -------
        None
        """
        # Setup tags; metadata (if any) and other key value pairs.
        tags = {"machine": f"{endpoint.name}"}
        tags.update(metadata)

        # The measurements to be posted to Influx
        top_lvl_names = predictions.columns.get_level_values(0).unique()

        for top_lvl_name in top_lvl_names:

            # Makes no sense to create a influx point where the field would be a date
            # and doesn't seem to be possible anyway.
            if top_lvl_name in ["end", "start"]:
                continue

            # this is a 'regular' non-stacked column dataframe
            sub_df = predictions[top_lvl_name]

            if isinstance(sub_df, pd.Series):
                sub_df = pd.DataFrame(sub_df)

            # Set the sub df's column names equal to the name of the tags if
            # they match the length of the tag list.
            if len(sub_df.columns) == len(endpoint.tag_list):
                sub_df.columns = [tag.name for tag in endpoint.tag_list]

            self._write_to_influx_with_retries(sub_df, tags, top_lvl_name)

    def _write_to_influx_with_retries(self, df, tags, measurement):
        """
        Write data to influx with retries and exponential backof. Will sleep
        exponentially longer between each retry, starting at 8 seconds, capped at 5 min.
        """
        logger.info(
            f"Writing {len(df)} points to Influx for measurement: {measurement}"
        )
        for current_attempt in itertools.count(start=1):
            try:
                self.dataframe_client.write_points(
                    dataframe=df, measurement=measurement, tags=tags, batch_size=10000
                )
            except Exception as exc:
                if current_attempt <= self.n_retries:
                    # Sleep at most 5 min
                    time_to_sleep = min(2 ** (current_attempt + 2), 300)
                    logger.warning(
                        f"Failed to forward data to influx on attempt "
                        f"{current_attempt} out of {self.n_retries}.\n"
                        f"Error: {exc}.\n"
                        f"Sleeping {time_to_sleep} seconds and trying again."
                    )
                    time.sleep(time_to_sleep)
                    continue
                else:
                    msg = f"Failed to forward data to influx. Error: {exc}"
                    logger.error(msg)
            else:
                break

    def send_sensor_data(self, sensors: pd.DataFrame):
        """
        Write sensor-data to influx
        """
        # Write the per-sensor points to influx
        logger.info(f"Writing {len(sensors)} sensor points to Influx")

        for tag_name, tag_data in _explode_df(sensors).items():
            self._write_to_influx_with_retries(tag_data, {"tag": tag_name}, "resampled")
            logger.debug(f"Wrote resampled tag data for {tag_name} to Influx")
        logger.debug("Done writing resampled sensor values to influx")


def _explode_df(
    df: pd.DataFrame, field_name: str = "value"
) -> typing.Dict[str, pd.DataFrame]:
    """
    Converts a single dataframe with several columns to a map from the column names
    to dataframes which has that single column as the only column with the name
    `field_name`

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame with one or more columns
    field_name:
        Name to give the data-columns in the resulting map

    Returns
    -------
    Dict: A map from column names to DataFrames containing only that column.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({"Column 1": [1,2,3], "Col2": [5,6,7]})
    >>> res = _explode_df(df)
    >>> len(res)
    2
    >>> sorted(list(res.keys()))
    ['Col2', 'Column 1']
    >>> res["Col2"]
       value
    0      5
    1      6
    2      7
    """
    ret = dict()
    for col in df.columns:
        ret[col] = df[[col]].rename(columns={col: field_name}, inplace=False)
    return ret



# -*- coding: utf-8 -*-

import os
import glob
import unittest
import logging
import nbformat
import tempfile
import dateutil.parser
import importlib
import sys

from unittest import mock

import pandas as pd
import numpy as np

from nbconvert.exporters import PythonExporter

from gordo_components.data_provider.providers import DataLakeProvider
from gordo_components.dataset.datasets import TimeSeriesDataset

logger = logging.getLogger(__name__)


def _fake_data():
    data = pd.DataFrame({f"sensor_{i}": np.random.random(size=100) for i in range(10)})
    return data, data


class ExampleNotebooksTestCase(unittest.TestCase):
    @mock.patch.object(TimeSeriesDataset, "get_data", return_value=_fake_data())
    def test_faked_DataLakeBackedDataset(self, _mocked_method):

        config = dict(
            from_ts=dateutil.parser.isoparse("2014-07-01T00:10:00+00:00"),
            to_ts=dateutil.parser.isoparse("2015-01-01T00:00:00+00:00"),
            tag_list=[
                "asgb.19ZT3950%2FY%2FPRIM",
                "asgb.19PST3925%2FDispMeasOut%2FPRIM",
            ],
        )

        provider = DataLakeProvider(storename="dataplatformdlsprod", interactive=True)
        dataset = TimeSeriesDataset(data_provider=provider, **config)

        # Should be able to call get_data without being asked to authenticate in tests
        X, y = dataset.get_data()

    @mock.patch.object(TimeSeriesDataset, "get_data", return_value=_fake_data())
    def test_notebooks(self, _mocked_method):
        """
        Ensures all notebooks will run without error
        """
        repo_root = os.path.join(os.path.dirname(__file__), "..")
        notebooks = glob.glob(os.path.join(repo_root, "examples", "*.ipynb"))

        logger.info(f"Found {len(notebooks)} notebooks to test.")

        for notebook in notebooks:

            logger.info(f"Testing notebook: {os.path.basename(notebook)}")

            with open(notebook) as f:
                nb = nbformat.read(f, as_version=4)
                exporter = PythonExporter()
                source, _meta = exporter.from_notebook_node(nb)

                with tempfile.TemporaryDirectory() as tmpdir:
                    with open(os.path.join(tmpdir, "tmpmodule.py"), "w") as f:
                        f.writelines(source)
                    with open(os.path.join(tmpdir, "__init__.py"), "w") as f:
                        f.write("from .tmpmodule import *")

                    # Import this module to 'run' the code
                    module_dir = os.path.join(tmpdir, "..")
                    sys.path.insert(0, module_dir)

                    importlib.import_module(os.path.basename(tmpdir), ".")

                    sys.path.remove(module_dir)



# -*- coding: utf-8 -*-

from typing import Dict, Any

import yaml
from gordo_components.workflow.config_elements.data_provider import DataProvider

from gordo_components.workflow.config_elements.dataset import Dataset
from gordo_components.workflow.config_elements.validators import (
    ValidUrlString,
    ValidMetadata,
    ValidModel,
    ValidDataset,
    ValidMachineRuntime,
    ValidDataProvider,
)
from gordo_components.workflow.workflow_generator.helpers import patch_dict
from .base import ConfigElement


class Machine(ConfigElement):
    """
    Represents a single machine in a config file
    """

    name = ValidUrlString()
    project_name = ValidUrlString()
    host = ValidUrlString()
    model = ValidModel()
    dataset = ValidDataset()
    data_provider = ValidDataProvider()
    metadata = ValidMetadata()
    runtime = ValidMachineRuntime()

    def __init__(
        self,
        name: str,
        model: dict,
        dataset: Dataset,
        data_provider: DataProvider,
        project_name: str,
        evaluation: dict,
        metadata=None,
        runtime=None,
    ):

        if runtime is None:
            runtime = dict()
        if metadata is None:
            metadata = dict()
        self.name = name
        self.model = model
        self.dataset = dataset
        self.data_provider = data_provider
        self.runtime = runtime
        self.evaluation = evaluation

        self.metadata = metadata
        self.project_name = project_name

        self.host = f"gordoserver-{self.project_name}-{self.name}"

    @classmethod
    def from_config(  # type: ignore
        cls, config: Dict[str, Any], project_name: str, config_globals=None
    ):
        if config_globals is None:
            config_globals = dict()

        name = config["name"]
        model = config.get("model") or config_globals.get("model")

        local_runtime = config.get("runtime", dict())
        runtime = patch_dict(config_globals.get("runtime", dict()), local_runtime)

        dataset_config = patch_dict(
            config.get("dataset", dict()), config_globals.get("dataset", dict())
        )
        dataset = Dataset.from_config(dataset_config)
        evaluation = patch_dict(
            config_globals.get("evaluation", dict()), config.get("evaluation", dict())
        )

        data_provider = DataProvider.from_config(
            patch_dict(
                config_globals.get("data_provider", dict()),
                config.get("data_provider", dict()),
            )
        )

        metadata = {
            "global-metadata": config_globals.get("metadata", dict()),
            "machine-metadata": config.get("metadata", dict()),
        }
        return cls(
            name,
            model,
            dataset,
            data_provider=data_provider,
            metadata=metadata,
            runtime=runtime,
            project_name=project_name,
            evaluation=evaluation,
        )

    def __str__(self):
        return yaml.dump(self.to_dict())

    def to_dict(self):
        return {
            "name": self.name,
            "dataset": self.dataset.to_dict(),
            "model": self.model,
            "metadata": self.metadata,
            "runtime": self.runtime,
            "data_provider": self.data_provider.to_dict(),
            "project_name": self.project_name,
            "evaluation": self.evaluation,
        }



# -*- coding: utf-8 -*-

import tempfile
import pathlib

from gordo_components.util import disk_registry


def test_simple_happy_path():
    """Tests that it works to write and read a simple value to a fresh registry"""
    with tempfile.TemporaryDirectory() as tmpdir:
        disk_registry.write_key(tmpdir, "akey", "aval")
        assert disk_registry.get_value(tmpdir, "akey") == "aval"


def test_new_registry():
    """Tests that it works to write and read a simple value to a fresh registry with
     a non-existing directory"""
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = pathlib.Path(tmpdir).joinpath("newregistry")
        disk_registry.write_key(registry, "akey", "aval")
        assert disk_registry.get_value(registry, "akey") == "aval"


def test_complicated_happy_path():
    """Tests that it works to write and read a 'complicated' value"""
    with tempfile.TemporaryDirectory() as tmpdir:
        value = """
        A long
        value with many weird character lie åøæ
        and some linebreaks"""
        disk_registry.write_key(tmpdir, "akey", value)
        assert disk_registry.get_value(tmpdir, "akey") == value


def test_overwrites_existing():
    """Double writes to the same registry overwrites the first value"""
    the_key = "akey"
    with tempfile.TemporaryDirectory() as tmpdir:
        first_value = "Some value"
        disk_registry.write_key(tmpdir, the_key, first_value)
        assert disk_registry.get_value(tmpdir, the_key) == first_value
        second_value = "Some value"
        disk_registry.write_key(tmpdir, the_key, second_value)
        assert disk_registry.get_value(tmpdir, the_key) == second_value


def test_delete():
    """Delete removes a key"""
    the_key = "akey"
    with tempfile.TemporaryDirectory() as tmpdir:
        first_value = "Some value"
        disk_registry.write_key(tmpdir, the_key, first_value)
        assert disk_registry.get_value(tmpdir, the_key) == first_value

        existed_p = disk_registry.delete_value(tmpdir, the_key)
        assert disk_registry.get_value(tmpdir, the_key) is None
        # They key existed
        assert existed_p


def test_double_delete():
    """Delete works on non-existing key, returning False"""
    the_key = "akey"
    with tempfile.TemporaryDirectory() as tmpdir:
        existed_p = disk_registry.delete_value(tmpdir, the_key)
        assert disk_registry.get_value(tmpdir, the_key) is None
        assert not existed_p


def test_get_value_without_registry_dir():
    """Test that it works to have registry_dir as None, and that it returns None as expected"""
    assert disk_registry.get_value(None, "akey") is None



# -*- coding: utf-8 -*-

import abc
import logging
from typing import Iterable, Union, List, Callable
from datetime import datetime

import pandas as pd
import numpy as np

logger = logging.getLogger(__name__)


class GordoBaseDataset:
    @abc.abstractmethod
    def get_data(self):
        """
        Using initialized params, returns X, y as numpy arrays.

        """

    @abc.abstractmethod
    def get_metadata(self):
        """
        Return metadata about the dataset in primitive / json encode-able dict form.
        """

    @staticmethod
    def join_timeseries(
        series_iterable: Iterable[pd.Series],
        resampling_startpoint: datetime,
        resampling_endpoint: datetime,
        resolution: str,
        aggregation_methods: Union[str, List[str], Callable] = "mean",
    ) -> pd.DataFrame:
        """

        Parameters
        ----------
        series_iterable: Iterable[pd.Series]
            An iterator supplying series with time index
        resampling_startpoint: datetime.datetime
            The starting point for resampling. Most data frames will not have this
            in their datetime index, and it will be inserted with a NaN as the value.
            The resulting NaNs will be removed, so the only important requirement for this is
            that this resampling_startpoint datetime must be before or equal to the first
            (earliest) datetime in the data to be resampled.
        resampling_endpoint: datetime.datetime
            The end point for resampling. This datetime must be equal to or after the last datetime in the
            data to be resampled.
        resolution: str
            The bucket size for grouping all incoming time data (e.g. "10T")
            Available strings come from https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        aggregation_methods: Union[str, List[str], Callable]
            Aggregation method(s) to use for the resampled buckets. If a single
            resample method is provided then the resulting dataframe will have names
            identical to the names of the series it got in. If several
            aggregation-methods are provided then the resulting dataframe will
            have a multi-level column index, with the series-name as the first level,
            and the aggregation method as the second level.
            See :py:func::`pandas.core.resample.Resampler#aggregate` for more
            information on possible aggregation methods.

        Returns
        -------
        pd.DataFrame
            A dataframe without NaNs, a common time index, and one column per
            element in the dataframe_generator. If multiple aggregation methods
            are provided then the resulting dataframe will have a multi-level column
            index with series-names as top-level and aggregation-method as second-level.

        """
        resampled_series = []

        for series in series_iterable:
            startpoint_sametz = resampling_startpoint.astimezone(
                tz=series.index[0].tzinfo
            )
            endpoint_sametz = resampling_endpoint.astimezone(tz=series.index[0].tzinfo)

            if series.index[0] > startpoint_sametz:
                # Insert a NaN at the startpoint, to make sure that all resampled
                # indexes are the same. This approach will "pad" most frames with
                # NaNs, that will be removed at the end.
                startpoint = pd.Series(
                    [np.NaN], index=[startpoint_sametz], name=series.name
                )
                series = startpoint.append(series)
                logging.debug(
                    f"Appending NaN to {series.name} " f"at time {startpoint_sametz}"
                )

            elif series.index[0] < resampling_startpoint:
                msg = (
                    f"Error - for {series.name}, first timestamp "
                    f"{series.index[0]} is before the resampling start point "
                    f"{startpoint_sametz}"
                )
                logging.error(msg)
                raise RuntimeError(msg)

            if series.index[-1] < endpoint_sametz:
                endpoint = pd.Series(
                    [np.NaN], index=[endpoint_sametz], name=series.name
                )
                series = series.append(endpoint)
                logging.debug(
                    f"Appending NaN to {series.name} " f"at time {endpoint_sametz}"
                )
            elif series.index[-1] > endpoint_sametz:
                msg = (
                    f"Error - for {series.name}, last timestamp "
                    f"{series.index[-1]} is later than the resampling end point "
                    f"{endpoint_sametz}"
                )
                logging.error(msg)
                raise RuntimeError(msg)

            logging.debug("Head (3) and tail(3) of dataframe to be resampled:")
            logging.debug(series.head(3))
            logging.debug(series.tail(3))

            resampled = series.resample(resolution, label="left").agg(
                aggregation_methods
            )
            # If several aggregation methods are provided, agg returns a dataframe
            # instead of a series. In this dataframe the column names are the
            # aggregation methods, like "max" and "mean", so we have to make a
            # multi-index with the series-name as the top-level and the
            # aggregation-method as the lower-level index.
            # For backwards-compatibility we *dont* return a multi-level index
            # when we have a single resampling method.
            if isinstance(
                resampled, pd.DataFrame
            ):  # Several aggregation methods provided
                resampled.columns = pd.MultiIndex.from_product(
                    [[series.name], resampled.columns],
                    names=["tag", "aggregation_method"],
                )
            filled = resampled.fillna(method="ffill")
            resampled_series.append(filled)

        new_series = pd.concat(resampled_series, axis=1, join="inner")
        # Before returning, delete all rows with NaN, they were introduced by the
        # insertion of NaNs in the beginning of all timeseries

        return new_series.dropna()



# -*- coding: utf-8 -*-
import _thread
import threading
import logging
import time
from typing import Callable, Dict, Optional

import kubernetes
from kubernetes import client as kubeclient, config, watch
from kubernetes.client.rest import ApiException


logger = logging.getLogger(__name__)


def load_config():
    """
    Attempts to load cluster config from $HOME/.kube/config with a fallback
    to the incluster
    Returns
    -------

    """
    try:
        config.load_kube_config()  # Local .kube/config
    except (FileNotFoundError, TypeError):
        config.load_incluster_config()  # Within cluster auth for a pod


class ThreadedWatcher(threading.Thread):
    """Thread watching for changes in kubernetes. Will restart on
    `kubernetes.client.rest.ApiException`, other exceptions causes an interrupt of the
    main thread.
    """

    def __init__(self, watched_function: Callable, event_handler: Callable, **kwargs):
        """
        A  thread watching for kubernetes changes using `watched_function`, invoking
        `event_handler` on each event emitted by the watched function.

        `watched_function` is expected to be of a type which can be used by
        :py:func:`kubernetes.watch.watch.Watch.stream`, most prominently
        functions in :py:class:`kubernetes.client.apis.core_v1_api.CoreV1Api`.

        The generated thread is returned, and it is the callers responsibility to call
        `start` on it. The thread is started in daemon-mode: the entire Python program
        exits when no alive non-daemon threads are left.

        If a unhandled exception occurs in this thread it will call
        :py:func:`_thread.interrupt_main`.



        Parameters
        ----------
        watched_function: Callable
            Function use to watch for kubernetes changes
        event_handler:
            Function which will be called to handle any events emitted by
            `watched_function`
        kwargs
            Extra arguments which will be passed in to `watched_function`

        """
        self.process_event = event_handler
        self.func = watched_function
        self.kwargs = kwargs
        threading.Thread.__init__(self, daemon=True)
        # If this is set then the threads dies after the next received element. Useful
        # for testing.
        self._die_after_next = False

    def run(self):
        try:
            while True:
                try:
                    w = watch.Watch()
                    for event in w.stream(self.func, **self.kwargs):
                        self.process_event(event)
                        if self._die_after_next:
                            return
                except ApiException:
                    logger.exception(
                        "Exception encountered while watching for event stream"
                    )
                    time.sleep(0.5)
                    pass
        except Exception:
            logger.exception("Got unhandled exception in k8s watching thread, exiting")
            _thread.interrupt_main()

    def die_after_next_elem(self, val=True):
        self._die_after_next = val


def watch_namespaced_custom_object(
    event_handler: Callable,
    namespace: str,
    client: Optional[kubernetes.client.apis.core_v1_api.CoreV1Api] = None,
    selectors: Optional[Dict[str, str]] = None,
) -> ThreadedWatcher:
    """Watches changes to k8s services in a given namespace, and executed
    `event_handler` for each event.

    Returns the watching thread, which must be started by the caller.

    Parameters
    ----------
    event_handler : Callable
        Function which will be called on each service-event from k8s. Should take a
        single argument, which will be the k8s event for a service change. See
        `kubernetes.watch.watch.Watch.stream` for description of event structure.
    namespace: str
        Namespace to look for changes in
    client: Optional[kubernetes.client.apis.core_v1_api.CoreV1Api]
        K8s client to use to watch for changes. If None then we try to create one.
    selectors: Optional[Dict[str, str]]


    Returns
    -------
    gordo_components.watchman.gordo_k8s_interface.ThreadedWatcher
        The watching thread, must be started by the caller.

    """

    if client is None:
        load_config()
        client = kubeclient.CustomObjectsApi()
    else:
        client = client

    kwargs = dict(
        event_handler=event_handler,
        watched_function=client.list_namespaced_custom_object,
        namespace=namespace,
        group="equinor.com",
        version="v1",
        plural="models",
    )
    if selectors:
        return ThreadedWatcher(
            label_selector=",".join(f"{k}={v}" for k, v in selectors.items()), **kwargs
        )
    else:
        return ThreadedWatcher(
            field_selector=f"metadata.namespace=={namespace}", **kwargs
        )



# -*- coding: utf-8 -*-

import os
import logging
from functools import wraps
from typing import List
from flask import Flask, jsonify, make_response, current_app
from flask.views import MethodView
from apscheduler.schedulers.background import BackgroundScheduler

from gordo_components import __version__
from gordo_components.watchman.endpoints_status import EndpointStatuses


logger = logging.getLogger(__name__)
logger.setLevel(os.environ.get("WATCHMAN_LOGLEVEL", "INFO").upper())


# Setup the scheduler to update the endpoint statuses
scheduler = BackgroundScheduler()

ENDPOINT_STATUSES = None  # Initialized below


class WatchmanApi(MethodView):
    """
    API view to list expected endpoints in this project space and report if they
    are up or not.
    """

    def get(self):

        payload = jsonify(
            {
                "endpoints": ENDPOINT_STATUSES.statuses(),
                "project-name": current_app.config["PROJECT_NAME"],
                "project-version": current_app.config["PROJECT_VERSION"],
            }
        )
        resp = make_response(payload, 200)
        resp.headers["Cache-Control"] = "max-age=5"
        return resp


def healthcheck():
    """
    Return gordo version, route for Watchman server
    """
    payload = jsonify(
        {
            "version": __version__,
            "config": current_app.config["TARGET_NAMES"],
            "project-name": current_app.config["PROJECT_NAME"],
            "project-version": current_app.config["PROJECT_VERSION"],
        }
    )
    return payload, 200


def run_override(func):
    """
    Wrapper to Flask's app.run function which will first start the scheduler
    and then call the original app.run()
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        scheduler.start()
        return func(*args, **kwargs)

    return wrapper


def build_app(
    project_name: str,
    project_version: str,
    target_names: List[str],
    namespace: str,
    ambassador_host: str,
    listen_to_kubernetes: bool = True,
):
    """
    Build app and any associated routes
    """

    global ENDPOINT_STATUSES
    ENDPOINT_STATUSES = EndpointStatuses(
        scheduler=scheduler,
        project_name=project_name,
        ambassador_host=ambassador_host,
        model_names=target_names,
        project_version=project_version,
        namespace=namespace,
        listen_to_kubernetes=listen_to_kubernetes,
    )

    # App and routes
    app = Flask(__name__)
    app.config.update(
        PROJECT_NAME=project_name,
        PROJECT_VERSION=project_version,
        TARGET_NAMES=list(target_names),
    )
    app.add_url_rule(rule="/healthcheck", view_func=healthcheck, methods=["GET"])
    app.add_url_rule(
        rule="/", view_func=WatchmanApi.as_view("watchman_api"), methods=["GET"]
    )

    # Ensure that calling app.run will start the scheduler
    app.run = run_override(app.run)  # type: ignore

    return app


def run_server(
    host: str,
    port: int,
    debug: bool,
    project_name: str,
    project_version: str,
    target_names: List[str],
    namespace: str,
    ambassador_host: str,
    listen_to_kubernetes: bool = True,
):
    app = build_app(
        project_name=project_name,
        project_version=project_version,
        target_names=target_names,
        namespace=namespace,
        ambassador_host=ambassador_host,
        listen_to_kubernetes=listen_to_kubernetes,
    )
    app.run(host, port, debug=debug, threaded=False)



import ast
import logging
import pandas as pd
import numpy as np

logger = logging.getLogger(__name__)


class _DfNameInserter(ast.NodeTransformer):
    def __init__(self, df_name="df"):
        """
        Constructs a ast.NodeTransformer which inserts df[X] around strings X.

        So if it is given
        ast.parse("('A' > 2) | (('B' < 10) & ('C' > 123.31))")
        it will return the same as
        ast.parse("(df['A'] > 2) | ((df['B'] < 10) & (df['C'] > 123.31))")

        Parameters
        ----------
        df_name : str
                  Name of the string to insert "around" strings.

        Examples
        --------
        >>> ast.dump(_DfNameInserter().visit(ast.parse("'A' > 2"))) == ast.dump(ast.parse("df['A'] > 2"))
        True

        """
        self.df_name = df_name
        super().__init__()

    def visit_Str(self, node):
        return ast.fix_missing_locations(
            ast.copy_location(
                ast.Subscript(
                    value=ast.Name(id=self.df_name, ctx=ast.Load()),
                    slice=ast.Index(value=ast.copy_location(ast.Str(s=node.s), node)),
                    ctx=ast.Load(),
                ),
                node,
            )
        )


def apply_buffer(mask: pd.Series, buffer_size: int = 0):
    """
    Take a mask (boolean series) where True indicates keeping a value, and False
    represents removing the value. This will 'expand' those indexes marked as `False`
    to the symmetrical bounds of ``buffer_size``

    Parameters
    ----------
    mask: pandas.core.Series
        Boolean pandas series
    buffer_size: int
        Size to buffer around ``False`` values

    Examples
    --------
    >>> import pandas as pd
    >>> series = pd.Series([True, True, False, True, True])
    >>> apply_buffer(series, buffer_size=1)
    >>> series
    0     True
    1    False
    2    False
    3    False
    4     True
    dtype: bool

    Returns
    -------
    None
    """
    idxs, *_rows = np.where(mask == False)
    for idx in idxs:
        mask.values[
            range(max((0, idx - buffer_size)), min((len(mask), idx + buffer_size + 1)))
        ] = False


def pandas_filter_rows(df, filter_str: str, buffer_size: int = 0):
    """

    Parameters
    ----------
    df: pandas.Dataframe
      Dataframe to filter rows from. Does not modify the parameter
    filter_str: str
      String representing the filter. Can be a boolean combination of conditions,
      where conditions are comparisons of column names and either other columns
      or numeric values. The rows matching the filter are kept.
      Column names can be quoted in single/double quotations or back ticks.
      Example of legal filters are " `Tag A` > 5 " , " ('Tag B' > 1) | ('Tag C' > 4)"
      '("Tag D" < 5) '
    buffer_size: int
      Area fore and aft of the application of ``fitler_str`` to also mark for removal.

    Returns
    -------
    pandas.Dataframe
        The dataframe containing only rows matching the filter

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> df = pd.DataFrame(list(np.ndindex((3,3))), columns=list('AB'))
    >>> df
       A  B
    0  0  0
    1  0  1
    2  0  2
    3  1  0
    4  1  1
    5  1  2
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "`A`>1")
       A  B
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "`A`>'B'")
       A  B
    3  1  0
    6  2  0
    7  2  1
    >>> pandas_filter_rows(df, "(`A`>1) | (`B`<1)")
       A  B
    0  0  0
    3  1  0
    6  2  0
    7  2  1
    8  2  2

    """
    filter_str = filter_str.replace("`", '"')
    parsed_filter = ast.parse(filter_str, mode="eval")
    if not _safe_ast(parsed_filter):
        raise ValueError(f"Unsafe expression {filter_str}")
    # Replaces strings (assumed to represent column names) with df[..]. df_name _must_
    # match the parameter name of this function
    parsed_filter = _DfNameInserter(df_name="df").visit(parsed_filter)
    # The eval requires the dataframe to be called 'df'
    pandas_filter = eval(compile(parsed_filter, filename=__name__, mode="eval"))
    apply_buffer(pandas_filter, buffer_size=buffer_size)
    return df[pandas_filter]


# Contains all the legal AST nodes. The parsed expression below contains all components
# allowed in a filter query.
_LEGAL_AST_NODES = {
    type(a)
    for a in ast.walk(  # type: ignore
        ast.parse(  # type: ignore
            '("A"<6) | ("B" <= 2) & ("C" == 4.3) | ("A">-0.1 ) | ~("A">= (1-2)) '
            '| ("A" < ("B"+1)) | ("A" < ("B" *1)) | ("A" < ("B"-1)) | ("A" < ("B"/2)) '
            '| ("A" < ("B" **2))',
            mode="eval",
        ).body
    )
}


def _safe_ast(some_ast):
    """
    Validates that the parameter AST only contain safe AST nodes, that is nodes which
    are deemed to be safe to evaluate. Those are the nodes in _LEGAL_AST_NODES,
    except that the top-level node is allowed to be _ast.Expression,


    Parameters
    ----------
    some_ast: a parsed

    Examples
    --------
    >>> _safe_ast(ast.parse('("A"<6) | ("B" <= 2)', mode="eval"))
    True
    >>> _safe_ast(ast.parse("sys.exit(0)", mode="eval"))
    False
    """
    logger.debug(f"Validating AST: {ast.dump(some_ast)}")
    if type(some_ast) is ast.Expression:
        some_ast = some_ast.body

    current_ast_nodes = {type(a) for a in list(ast.walk(some_ast))}
    is_subset = current_ast_nodes.issubset(_LEGAL_AST_NODES)
    if not is_subset:
        logger.debug(
            f"Found that AST contained illegal nodes: "
            f"{current_ast_nodes.difference(_LEGAL_AST_NODES)}"
        )
    return is_subset



# -*- coding: utf-8 -*-

import unittest

import pytest
import yaml
import pandas as pd
import numpy as np

from gordo_components.model.transformers.imputer import InfImputer
from gordo_components import serializer

from sklearn.preprocessing import FunctionTransformer
from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA


class GordoFunctionTransformerFuncsTestCase(unittest.TestCase):
    """
    Test all functions within gordo_components meants for use in a Scikit-Learn
    FunctionTransformer work as expected
    """

    def _validate_transformer(self, transformer):
        """
        Inserts a transformer into the middle of a pipeline and runs it
        """
        pipe = Pipeline([("pca1", PCA()), ("custom", transformer), ("pca2", PCA())])
        X = np.random.random(size=100).reshape(10, 10)
        pipe.fit_transform(X)

    def test_multiply_by_function_transformer(self):
        from gordo_components.model.transformer_funcs.general import multiply_by

        # Provide a require argument
        tf = FunctionTransformer(func=multiply_by, kw_args={"factor": 2})
        self._validate_transformer(tf)

        # Ignore the required argument
        tf = FunctionTransformer(func=multiply_by)
        with self.assertRaises(TypeError):
            self._validate_transformer(tf)


@pytest.mark.parametrize("strategy", ["extremes", "minmax"])
def test_infimputer_basic(strategy):
    """
    Functionality of the InfImputer
    """
    base_x = np.random.random((100, 10)).astype(np.float32)

    flat_view = base_x.ravel()

    pos_inf_idxs = np.random.randint(0, len(flat_view), size=100)
    neg_inf_idxs = np.random.randint(0, len(flat_view), size=100)

    flat_view[pos_inf_idxs] = np.inf
    flat_view[neg_inf_idxs] = -np.inf

    # Our base x should now be littered with pos/neg inf values
    assert np.isposinf(base_x).sum() > 0, "Expected some positive infinity values here"
    assert np.isneginf(base_x).sum() > 0, "Expected some negative infinity values here"

    imputer = InfImputer(strategy=strategy, delta=2.0)

    # Test imputer on numpy array
    X = base_x.copy()
    X = imputer.fit_transform(X)
    assert np.isposinf(X).sum() == 0, "Expected no positive infinity values here"
    assert np.isneginf(X).sum() == 0, "Expected no negative infinity values here"

    if strategy == "extremes":
        # All pos infs in base_x should be filled in X now as the max dtype value
        assert np.all(X[np.where(np.isposinf(base_x))] == np.finfo(X.dtype).max)

        # All neg infs in base_x should be filled in X now as the min dtype value
        assert np.all(X[np.where(np.isneginf(base_x))] == np.finfo(X.dtype).min)

    # min max is a bit more difficult to assert however.
    elif strategy == "minmax":

        # Identify the features in the base array which have pos infs
        features_with_pos_infs = np.where(np.isposinf(base_x))[1]

        # Get the maxes of those features in the imputed X
        filled_maxes = X[:, features_with_pos_infs].max(axis=0)

        # Get the previous maxes
        previous_maxes = np.ma.masked_invalid(base_x[:, features_with_pos_infs]).max(
            axis=0
        )

        # Compare that each new max is the previous max + 2.
        # These were previously 'inf' values, replaced by previous max + 2
        for filled_max, previous_max in zip(filled_maxes, previous_maxes):
            assert np.isclose(filled_max, previous_max + 2.0)

        ### Repeat the process for negative infs ###

        # Identify the features in the base array which have pos infs
        features_with_neg_infs = np.where(np.isneginf(base_x))[1]

        # Get the maxes of those features in the imputed X
        filled_mins = X[:, features_with_neg_infs].min(axis=0)

        # Get the previous maxes
        previous_mins = np.ma.masked_invalid(base_x[:, features_with_neg_infs]).min(
            axis=0
        )

        # Now compare, that each new min is the min - 2. wq
        for filled_min, previous_min in zip(filled_mins, previous_mins):
            assert np.isclose(filled_min, previous_min - 2.0)

    # Test imputer on pandas dataframe
    X = pd.DataFrame(base_x.copy())
    X = imputer.fit_transform(X)
    assert np.isposinf(X).sum() == 0, "Expected no positive infinity values here"
    assert np.isneginf(X).sum() == 0, "Expected no negative infinity values here"


def test_infimputer_fill_values():
    """
    InfImputer when fill values are provided
    """
    base_x = np.random.random((100, 10)).astype(np.float32)

    flat_view = base_x.ravel()

    pos_inf_idxs = [1, 2, 3, 4, 5]
    neg_inf_idxs = [6, 7, 8, 9, 10]

    flat_view[pos_inf_idxs] = np.inf
    flat_view[neg_inf_idxs] = -np.inf

    # Our base x should now be littered with pos/neg inf values
    assert np.isposinf(base_x).sum() > 0, "Expected some positive infinity values here"
    assert np.isneginf(base_x).sum() > 0, "Expected some negative infinity values here"

    imputer = InfImputer(inf_fill_value=9999.0, neg_inf_fill_value=-9999.0)
    X = imputer.fit_transform(base_x)
    np.equal(
        X.ravel()[[pos_inf_idxs]], np.array([9999.0, 9999.0, 9999.0, 9999.0, 9999.0])
    )
    np.equal(
        X.ravel()[[neg_inf_idxs]],
        np.array([-9999.0, -9999.0, -9999.0, -9999.0, -9999.0]),
    )


@pytest.mark.parametrize(
    "config_str",
    [
        """
    sklearn.pipeline.Pipeline:
      steps:
        - gordo_components.model.transformers.imputer.InfImputer
    """,
        """
    sklearn.pipeline.Pipeline:
      steps:
        - gordo_components.model.transformers.imputer.InfImputer:
            inf_fill_value: 10
    """,
        """gordo_components.model.transformers.imputer.InfImputer""",
    ],
)
def test_imputer_from_definition(config_str: str):
    """
    Ensure it plays well with the gordo serializer
    """
    config = yaml.safe_load(config_str)
    model = serializer.pipeline_from_definition(config)

    if isinstance(model, Pipeline):
        assert isinstance(model.steps[-1][1], InfImputer)
    else:
        assert isinstance(model, InfImputer)

    serializer.pipeline_from_definition(serializer.pipeline_into_definition(model))



import logging
import dateutil.parser

import pytest

from gordo_components.data_provider.providers import InfluxDataProvider
from gordo_components.client.utils import influx_client_from_uri
from gordo_components.dataset.dataset import _get_dataset

from tests import utils as tu


logger = logging.getLogger(__name__)


def test_read_single_sensor_empty_data_time_range_indexerror(influxdb, caplog):
    """
    Asserts that an IndexError is raised because the dates requested are outside the existing time period
    """
    from_ts = dateutil.parser.isoparse("2017-01-01T09:11:00+00:00")
    to_ts = dateutil.parser.isoparse("2017-01-01T10:30:00+00:00")

    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=tu.INFLUXDB_URI, dataframe_client=True),
    )

    with caplog.at_level(logging.CRITICAL):
        with pytest.raises(IndexError):
            ds.read_single_sensor(
                from_ts=from_ts,
                to_ts=to_ts,
                tag=tu.SENSORS_STR_LIST[0],
                measurement="sensors",
            )


def test_read_single_sensor_empty_data_invalid_tag_name_valueerror(influxdb):
    """
    Asserts that a ValueError is raised because the tag name inputted is invalid
    """
    from_ts = dateutil.parser.isoparse("2016-01-01T09:11:00+00:00")
    to_ts = dateutil.parser.isoparse("2016-01-01T10:30:00+00:00")

    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=tu.INFLUXDB_URI, dataframe_client=True),
    )
    with pytest.raises(ValueError):
        ds.read_single_sensor(
            from_ts=from_ts,
            to_ts=to_ts,
            tag="tag-does-not-exist",
            measurement="sensors",
        )


def test__list_of_tags_from_influx_validate_tag_names(influxdb):
    """
    Test expected tags in influx match the ones actually in influx.
    """
    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=tu.INFLUXDB_URI, dataframe_client=True),
    )
    list_of_tags = ds._list_of_tags_from_influx()
    expected_tags = tu.SENSORS_STR_LIST
    tags = set(list_of_tags)
    assert set(expected_tags) == tags, (
        f"Expected tags = {expected_tags}" f"outputted {tags}"
    )


def test_get_list_of_tags(influxdb):
    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=tu.INFLUXDB_URI, dataframe_client=True),
    )
    expected_tags = set(tu.SENSORS_STR_LIST)

    tags = set(ds.get_list_of_tags())
    assert expected_tags == tags

    # The cache does not screw stuff up
    tags = set(ds.get_list_of_tags())
    assert expected_tags == tags


def test_influx_dataset_attrs(influxdb):
    """
    Test expected attributes
    """
    from_ts = dateutil.parser.isoparse("2016-01-01T09:11:00+00:00")
    to_ts = dateutil.parser.isoparse("2016-01-01T10:30:00+00:00")
    tag_list = tu.SENSORTAG_LIST
    config = {
        "type": "TimeSeriesDataset",
        "from_ts": from_ts,
        "to_ts": to_ts,
        "tag_list": tag_list,
    }
    config["data_provider"] = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=tu.INFLUXDB_URI, dataframe_client=True),
    )
    dataset = _get_dataset(config)
    assert hasattr(dataset, "get_metadata")

    metadata = dataset.get_metadata()
    assert isinstance(metadata, dict)


def test_influx_load_series_dry_run_raises():
    ds = InfluxDataProvider(measurement="sensors", value_name="Value", client=None)
    from_ts = dateutil.parser.isoparse("2016-01-01T09:11:00+00:00")
    to_ts = dateutil.parser.isoparse("2016-01-01T10:30:00+00:00")
    tag_list = tu.SENSORTAG_LIST
    with pytest.raises(NotImplementedError):
        ds.load_series(from_ts=from_ts, to_ts=to_ts, tag_list=tag_list, dry_run=True)



# -*- coding: utf-8 -*-

from datetime import timedelta

import pytest
import numpy as np
import pandas as pd
import yaml

from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler, RobustScaler

from gordo_components import serializer
from gordo_components.model import utils as model_utils
from gordo_components.model.anomaly import DiffBasedAnomalyDetector
from gordo_components.model.anomaly.base import AnomalyDetectorBase


@pytest.mark.parametrize("scaler", (MinMaxScaler(), RobustScaler()))
@pytest.mark.parametrize(
    "index", (range(10), pd.date_range("2019-01-01", "2019-01-30", periods=10))
)
@pytest.mark.parametrize("lookback", (0, 5))
def test_diff_detector(scaler, index, lookback):
    """
    Test the functionality of the DiffBasedAnomalyDetector
    """

    # Some dataset.
    X, y = (
        pd.DataFrame(np.random.random((10, 3))),
        pd.DataFrame(np.random.random((10, 3))),
    )

    # Use PCA because it auto adjusts to model input/output like our gordo models
    base_estimator = PCA()

    # Give PCA a .predict method for the anomaly detector
    base_estimator.predict = base_estimator.transform

    model = DiffBasedAnomalyDetector(base_estimator=base_estimator, scaler=scaler)

    assert isinstance(model, AnomalyDetectorBase)

    assert model.get_params() == dict(base_estimator=base_estimator, scaler=scaler)

    model.fit(X, y)

    output: np.ndarray = model.predict(X)
    base_df = model_utils.make_base_dataframe(
        tags=["A", "B", "C"], model_input=X, model_output=output, index=index
    )

    # Base prediction dataframe has none of these columns
    assert not any(col in base_df.columns for col in ("total-anomaly", "tag-anomaly"))

    # Apply the anomaly detection logic on the base prediction df
    anomaly_df = model.anomaly(X, y, timedelta(days=1))

    # Should have these added error calculated columns now.
    assert all(col in anomaly_df.columns for col in ("total-anomaly", "tag-anomaly"))

    # Verify calculations
    feature_wise_error = np.abs(
        scaler.transform(base_df["model-output"].values) - scaler.transform(y)
    )
    assert np.allclose(feature_wise_error, anomaly_df["tag-anomaly"].values)

    total_anomaly = np.linalg.norm(feature_wise_error, axis=1)
    assert np.allclose(total_anomaly, anomaly_df["total-anomaly"].values)


@pytest.mark.parametrize(
    "config",
    (
        """
        gordo_components.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
                sklearn.pipeline.Pipeline:
                    steps:
                        - sklearn.preprocessing.data.MinMaxScaler
                        - gordo_components.model.models.KerasAutoEncoder:
                            kind: feedforward_hourglass
                    memory:
    """,
        """
        gordo_components.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
                gordo_components.model.models.KerasAutoEncoder:
                    kind: feedforward_hourglass
    """,
    ),
)
def test_diff_detector_serializability(config):
    """
    Should play well with the gordo serializer
    """
    config = yaml.load(config)

    model = serializer.pipeline_from_definition(config)
    serializer.pipeline_into_definition(model)
    serialized_bytes = serializer.dumps(model)
    serializer.loads(serialized_bytes)



import logging

import time
from mock import MagicMock

from gordo_components.watchman.gordo_k8s_interface import watch_namespaced_custom_object

logger = logging.getLogger(__name__)


def test_watch_service(mocker):
    mock_client = MagicMock()

    last_seen_value = None

    def assign_to_last_sendt_var(x):
        nonlocal last_seen_value
        last_seen_value = x

    called_count = 0

    def return_increasing_nr_and_sleep(**_kwargs):
        nonlocal called_count
        time.sleep(called_count)
        called_count = called_count + 1
        return called_count

    def watch_stream_mock(_self, func, **_kwargs):
        return [func()]

    mocker.patch("kubernetes.watch.watch.Watch.stream", watch_stream_mock)
    mock_client.list_service_for_all_namespaces = return_increasing_nr_and_sleep

    watch_thread = watch_namespaced_custom_object(
        event_handler=assign_to_last_sendt_var,
        namespace="somenamespace",
        client=mock_client,
    )
    # Ensure watch_thread dies after its first element so we dont have dangling threads
    # laying around
    watch_thread.die_after_next_elem()
    watch_thread.start()
    # Give the background thread some time to run, up to 3 seconds.
    max_ms_to_wait = 3000
    while last_seen_value is None and max_ms_to_wait > 0:
        time.sleep(0.01)
        max_ms_to_wait = max_ms_to_wait - 1
    assert last_seen_value is not None
    # Ensure watch_thread is done
    watch_thread.join()

    # And again, but this time with selectors
    watch_thread = watch_namespaced_custom_object(
        event_handler=assign_to_last_sendt_var,
        namespace="somenamespace",
        client=mock_client,
        selectors={
            "app.kubernetes.io/component": "service",
            "app.kubernetes.io/managed-by": "gordo",
        },
    )
    # Ensure watch_thread dies after its first element so we dont have dangling threads
    # laying around
    watch_thread.die_after_next_elem()
    watch_thread.start()
    # Give the background thread some time to run, up to 3 seconds.
    max_ms_to_wait = 3000
    while last_seen_value is None and max_ms_to_wait > 0:
        time.sleep(0.01)
        max_ms_to_wait = max_ms_to_wait - 1
    assert last_seen_value is not None
    # Ensure watch_thread is done
    watch_thread.join()



from typing import List, Dict, Optional

import requests
import logging


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

from playhouse.postgres_ext import (
    Model,
    PostgresqlExtDatabase,
    CharField,
    BinaryJSONField,
    DateTimeTZField,
)

db = PostgresqlExtDatabase(None)


class Machine(Model):
    name = CharField(index=True, unique=True)
    metadata = BinaryJSONField()
    train_start_date = DateTimeTZField()
    train_end_date = DateTimeTZField()

    class Meta:
        database = db

    @classmethod
    def from_endpoint(cls, ep):
        """
        Returns a Machine object constructed from a valid endpoint.

        Parameters
        ----------
        ep: Dict
            Endpoint from watchman

        Returns
        -------
        Machine

        """
        return cls(**endpoint_to_machine_data(ep))

    def __repr__(self):
        return f"Machine {self.__data__} "


def endpoint_to_machine_data(ep):
    name = ep["endpoint-metadata"]["metadata"]["name"]
    train_start_date = ep["endpoint-metadata"]["metadata"]["dataset"][
        "train_start_date"
    ]
    train_end_date = ep["endpoint-metadata"]["metadata"]["dataset"]["train_end_date"]
    return dict(
        name=name,
        metadata=ep,
        train_start_date=train_start_date,
        train_end_date=train_end_date,
    )


def watchman_to_sql(
    watchman_address: str,
    sql_host: str,
    sql_port: int = 5432,
    sql_database: str = "postgres",
    sql_username: str = "postgres",
    sql_password: str = None,
):
    """
    Connects to watchman, downloads all metadata, and proceeds to push them to postgres.

    Parameters
    ----------
    watchman_address: str
        Full URL (including http/https) to watchman
    sql_host: str
        Host where postgres resides, e.g. "localhost"
    sql_port: int
        Port for postgres
    sql_database: str
        Database name the metadata storage
    sql_username: str
        Username to use for authenticating towards postgres
    sql_password: str
        Password to use for authenticating towards postgres

    Returns
    -------
    bool
        Returns True if we successfully got all metadata for all machines, False
        otherwise

    """
    sql_parameters = {"host": sql_host, "port": sql_port, "user": sql_username}
    if sql_password:
        sql_parameters.update({"password": sql_password})
    db.init(sql_database, **sql_parameters)
    Machine.create_table(safe=True)
    machines = get_machines_from_watchman(watchman_address)
    got_all_machines = None not in machines
    logger.debug(f"Fetched a total of {len(machines)} machines")
    machines = [machine for machine in machines if machine is not None]
    logger.debug(f"Fetched a total of {len(machines)} healthy machines")
    with db.atomic():
        for machine in machines:
            logger.debug(f"Inserting machine {machine['name']} in sql")  # type: ignore
            Machine.insert(machine).on_conflict_ignore().execute()
    return got_all_machines


def get_machines_from_watchman(watchman_address: str) -> List[Optional[Machine]]:
    """
    Returns a list of `Machine` objects from watchman, with None indicating a failed
    machine.

    Parameters
    ----------
    watchman_address: str
        Full address of watchman, including "http" / "https"

    Returns
    -------
    List[Optional[Machine]]
        List of Machine objects, with possible None values indicating a failed machine.

    """
    ret: List[Optional[Machine]] = []
    response = requests.get(watchman_address, timeout=5)
    if response.ok:
        watchman_response = response.json()
        ret = _extract_machines_from_watchman_response(watchman_response)
    return ret


def _extract_machines_from_watchman_response(
    watchman_response: Dict,
) -> List[Optional[Machine]]:
    """
    Extracts the list of machines from the watchman json response, with None indicating
    a failed machine.

    Parameters
    ----------
    watchman_response: Dict
        Response from watchman

    Examples
    --------
    >>> example_response = {
    ...     "endpoints": [
    ...         {
    ...             "endpoint-metadata": {
    ...                    "metadata": {
    ...                       "user-defined": {"machine-name": "test_machine1"},
    ...                        "dataset": {
    ...                            "train_start_date": "2019-05-08T14:10:38Z",
    ...                            "train_end_date": "2019-06-08T14:10:38Z",
    ...                        },
    ...                        "name": "machine-1"
    ...                    }},
    ...             "healthy": False
    ...          },
    ...          {
    ...             "endpoint-metadata": {
    ...                    "metadata": {
    ...                       "user-defined": {"machine-name": "test_machine2"},
    ...                        "dataset": {
    ...                            "train_start_date": "2019-05-08T14:10:38Z",
    ...                            "train_end_date": "2019-06-08T14:10:38Z",
    ...                        },
    ...                        "name": "machine-2"
    ...                    }},
    ...             "healthy": True
    ...          }
    ...     ]
    ... }
    >>> machines = _extract_machines_from_watchman_response(example_response)
    >>> len(machines)
    2
    >>> len([m for m in machines if m is not None] )
    1

    Returns
    -------
    List[Optional[Machine]]
        List of Machine objects, with possible None values indicating a failed machine.
    """
    machines = []
    if "endpoints" in watchman_response:
        for ep in watchman_response["endpoints"]:
            if ep["healthy"]:
                machines.append(endpoint_to_machine_data(ep))
            else:
                logger.info(f"Found non-healthy endpoint {ep}, ignoring")
                machines.append(None)
    return machines



# -*- coding: utf-8 -*-

import unittest
from typing import List, Iterable, Optional

import pytest
import numpy as np
import pandas as pd
import dateutil.parser
from datetime import datetime

from gordo_components.data_provider.base import GordoBaseDataProvider
from gordo_components.dataset.datasets import RandomDataset, TimeSeriesDataset
from gordo_components.dataset.base import GordoBaseDataset
from gordo_components.dataset.sensor_tag import SensorTag


class DatasetTestCase(unittest.TestCase):
    def test_random_dataset_attrs(self):
        """
        Test expected attributes
        """

        start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
        end = dateutil.parser.isoparse("2017-12-29 06:00:00Z")

        dataset = RandomDataset(
            from_ts=start,
            to_ts=end,
            tag_list=[SensorTag("Tag 1", None), SensorTag("Tag 2", None)],
        )

        self.assertTrue(isinstance(dataset, GordoBaseDataset))
        self.assertTrue(hasattr(dataset, "get_data"))
        self.assertTrue(hasattr(dataset, "get_metadata"))

        X, y = dataset.get_data()
        self.assertTrue(isinstance(X, pd.DataFrame))

        # y can either be None or an numpy array
        self.assertTrue(isinstance(y, pd.DataFrame) or y is None)

        metadata = dataset.get_metadata()
        self.assertTrue(isinstance(metadata, dict))

    def test_join_timeseries(self):
        timeseries_list, latest_start, earliest_end = self.create_timeseries_list()

        self.assertTrue(
            len(timeseries_list[0]) > len(timeseries_list[1]) > len(timeseries_list[2])
        )

        frequency = "7T"
        timedelta = pd.Timedelta("7 minutes")
        resampling_start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
        resampling_end = dateutil.parser.isoparse("2018-01-15 08:00:00Z")
        all_in_frame = GordoBaseDataset.join_timeseries(
            timeseries_list, resampling_start, resampling_end, frequency
        )

        # Check that first resulting resampled, joined row is within "frequency" from
        # the real first data point
        self.assertGreaterEqual(
            all_in_frame.index[0], pd.Timestamp(latest_start) - timedelta
        )
        self.assertLessEqual(all_in_frame.index[-1], pd.Timestamp(resampling_end))

    def test_join_timeseries_nonutcstart(self):
        timeseries_list, latest_start, earliest_end = self.create_timeseries_list()
        frequency = "7T"
        resampling_start = dateutil.parser.isoparse("2017-12-25 06:00:00+07:00")
        resampling_end = dateutil.parser.isoparse("2018-01-12 13:07:00+07:00")
        all_in_frame = GordoBaseDataset.join_timeseries(
            timeseries_list, resampling_start, resampling_end, frequency
        )
        self.assertEqual(len(all_in_frame), 1854)

    def test_join_timeseries_with_gaps(self):
        timeseries_list, latest_start, earliest_end = self.create_timeseries_list()

        self.assertTrue(
            len(timeseries_list[0]) > len(timeseries_list[1]) > len(timeseries_list[2])
        )

        timeseries_with_holes = self.remove_data(
            timeseries_list,
            remove_from="2018-01-03 10:00:00Z",
            remove_to="2018-01-03 18:00:00Z",
        )

        frequency = "10T"
        resampling_start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
        resampling_end = dateutil.parser.isoparse("2018-01-12 07:00:00Z")

        all_in_frame = GordoBaseDataset.join_timeseries(
            timeseries_with_holes, resampling_start, resampling_end, frequency
        )
        self.assertEqual(all_in_frame.index[0], pd.Timestamp(latest_start))
        self.assertEqual(all_in_frame.index[-1], pd.Timestamp(resampling_end))

        expected_index = pd.date_range(
            start=dateutil.parser.isoparse(latest_start),
            end=resampling_end,
            freq=frequency,
        )
        self.assertListEqual(list(all_in_frame.index), list(expected_index))

    @staticmethod
    def remove_data(timeseries_list, remove_from, remove_to):
        timeseries_with_holes = []
        for timeseries in timeseries_list:
            timeseries = timeseries[
                (timeseries.index < remove_from) | (timeseries.index >= remove_to)
            ]
            timeseries_with_holes.append(timeseries)

        return timeseries_with_holes

    @staticmethod
    def create_timeseries_list():
        # Create three dataframes with different resolution and different start/ends
        # Test for no NaNs, test for correct first and last date
        latest_start = "2018-01-03 06:00:00Z"
        earliest_end = "2018-01-05 06:00:00Z"

        index_seconds = pd.date_range(
            start="2018-01-01 06:00:00Z", end="2018-01-07 06:00:00Z", freq="S"
        )
        index_minutes = pd.date_range(
            start="2017-12-28 06:00:00Z", end=earliest_end, freq="T"
        )
        index_hours = pd.date_range(
            start=latest_start, end="2018-01-12 06:00:00Z", freq="H"
        )

        timeseries_seconds = pd.Series(
            data=np.random.randint(0, 100, len(index_seconds)),
            index=index_seconds,
            name="ts-seconds",
        )
        timeseries_minutes = pd.Series(
            data=np.random.randint(0, 100, len(index_minutes)),
            index=index_minutes,
            name="ts-minutes",
        )
        timeseries_hours = pd.Series(
            data=np.random.randint(0, 100, len(index_hours)),
            index=index_hours,
            name="ts-hours",
        )

        return (
            [timeseries_seconds, timeseries_minutes, timeseries_hours],
            latest_start,
            earliest_end,
        )


class TimeSeriesDatasetTest(unittest.TestCase):
    """
    Tests the TimeSeriesDataset implementation with a mock datasource
    """

    def test_row_filter(self):
        """Tests that row_filter filters away rows"""
        kwargs = dict(
            data_provider=MockDataSource(),
            tag_list=[
                SensorTag("Tag 1", None),
                SensorTag("Tag 2", None),
                SensorTag("Tag 3", None),
            ],
            from_ts=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
            to_ts=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
        )
        X, _ = TimeSeriesDataset(**kwargs).get_data()
        self.assertEqual(577, len(X))

        X, _ = TimeSeriesDataset(row_filter="'Tag 1' < 5000", **kwargs).get_data()
        self.assertEqual(8, len(X))

        X, _ = TimeSeriesDataset(
            row_filter="'Tag 1' / 'Tag 3' < 0.999", **kwargs
        ).get_data()
        self.assertEqual(3, len(X))

    def test_aggregation_methods(self):
        """Tests that it works to set aggregation method(s)"""

        kwargs = dict(
            data_provider=MockDataSource(),
            tag_list=[
                SensorTag("Tag 1", None),
                SensorTag("Tag 2", None),
                SensorTag("Tag 3", None),
            ],
            from_ts=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
            to_ts=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
        )

        # Default aggregation gives no extra columns
        X, _ = TimeSeriesDataset(**kwargs).get_data()

        self.assertEqual((577, 3), X.shape)
        # The default single aggregation method gives the tag-names as columns
        self.assertEqual(list(X.columns), ["Tag 1", "Tag 2", "Tag 3"])

        # Using two aggregation methods give a multi-level column with tag-names
        # on top and aggregation_method as second level
        X, _ = TimeSeriesDataset(
            aggregation_methods=["mean", "max"], **kwargs
        ).get_data()

        self.assertEqual((577, 6), X.shape)
        self.assertEqual(
            list(X.columns),
            [
                ("Tag 1", "mean"),
                ("Tag 1", "max"),
                ("Tag 2", "mean"),
                ("Tag 2", "max"),
                ("Tag 3", "mean"),
                ("Tag 3", "max"),
            ],
        )


def test_time_series_no_resolution():
    kwargs = dict(
        data_provider=MockDataSource(),
        tag_list=[
            SensorTag("Tag 1", None),
            SensorTag("Tag 2", None),
            SensorTag("Tag 3", None),
        ],
        from_ts=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
        to_ts=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
    )
    no_resolution, _ = TimeSeriesDataset(resolution=None, **kwargs).get_data()
    wi_resolution, _ = TimeSeriesDataset(resolution="10T", **kwargs).get_data()
    assert len(no_resolution) > len(wi_resolution)


@pytest.mark.parametrize(
    "tag_list",
    [
        [SensorTag("Tag 1", None), SensorTag("Tag 2", None), SensorTag("Tag 3", None)],
        [SensorTag("Tag 1", None)],
    ],
)
@pytest.mark.parametrize(
    "target_tag_list",
    [
        [SensorTag("Tag 2", None), SensorTag("Tag 1", None), SensorTag("Tag 3", None)],
        [SensorTag("Tag 1", None)],
        [SensorTag("Tag10", None)],
        [],
    ],
)
def test_timeseries_target_tags(tag_list, target_tag_list):
    start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
    end = dateutil.parser.isoparse("2017-12-29 06:00:00Z")
    tsd = TimeSeriesDataset(
        start,
        end,
        tag_list=tag_list,
        target_tag_list=target_tag_list,
        data_provider=MockDataSource(),
    )
    X, y = tsd.get_data()

    # If we have targets, y and X should be equal length and axis=1 == N in target tag list
    if target_tag_list:
        assert len(X) == len(y)
        assert y.shape[1] == len(target_tag_list)

        # Ensure the order in maintained
        assert [tag.name for tag in target_tag_list] == y.columns.tolist()
    else:
        assert y is None

    # Features should match the tag_list
    assert X.shape[1] == len(tag_list)

    # Ensure the order in maintained
    assert [tag.name for tag in tag_list] == X.columns.tolist()


class MockDataSource(GordoBaseDataProvider):
    def __init__(self, **kwargs):
        pass

    def can_handle_tag(self, tag):
        return True

    def load_series(
        self,
        from_ts: datetime,
        to_ts: datetime,
        tag_list: List[SensorTag],
        dry_run: Optional[bool] = False,
    ) -> Iterable[pd.Series]:
        days = pd.date_range(from_ts, to_ts, freq="s")
        tag_list_strings = sorted([tag.name for tag in tag_list])
        for i, name in enumerate(tag_list_strings):
            series = pd.Series(
                index=days, data=list(range(i, len(days) + i)), name=name
            )
            yield series


def test_timeseries_dataset_compat():
    """
    There are accepted keywords in the config file when using type: TimeSeriesDataset
    which don't actually match the kwargs of the dataset's __init__; for compatibility
    :func:`gordo_components.dataset.datasets.compat` should adjust for these differences.
    """
    dataset = TimeSeriesDataset(
        data_provider=MockDataSource(),
        train_start_date="2017-12-25 06:00:00Z",
        train_end_date="2017-12-29 06:00:00Z",
        tags=[SensorTag("Tag 1", None)],
    )
    assert dataset.from_ts == dateutil.parser.isoparse("2017-12-25 06:00:00Z")
    assert dataset.to_ts == dateutil.parser.isoparse("2017-12-29 06:00:00Z")
    assert dataset.tag_list == [SensorTag("Tag 1", None)]



from setuptools import setup, find_packages


setup_requirements = ["pytest-runner", "setuptools_scm"]


def requirements(fp: str):
    with open(fp) as f:
        return [r.strip() for r in f.readlines()]


setup(
    author="Equinor ASA",
    author_email="fg_gpl@equinor.com",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Train and build models for Argo / Kubernetes",
    entry_points={"console_scripts": ["gordo-components=gordo_components.cli:gordo"]},
    install_requires=requirements("requirements.in"),  # Allow flexible deps for install
    license="AGPLv3",
    name="gordo-components",
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=requirements("test_requirements.txt"),
    url="https://github.com/equinor/gordo-components",
    use_scm_version={
        "write_to": "gordo_components/_version.py",
        "relative_to": __file__,
    },
    zip_safe=True,
    package_data={
        "": [
            "python/gordo_components/workflow/workflow_generator/resources/argo-workflow.yml.template"
        ]
    },
    include_package_data=True,
)



# -*- config: utf-8 -*-

import abc
from typing import Dict


class ConfigElement(abc.ABC):
    """
    Represents functionality of an element found within a gordo config file.
    """

    @abc.abstractclassmethod
    def from_config(cls, config: Dict[str, dict]) -> "ConfigElement":
        """
        Take one section of a config file and parse it into a given element
        """
        ...



# -*- coding: utf-8 -*-

import unittest
import logging

from tempfile import TemporaryDirectory

import pytest
import numpy as np
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.decomposition import TruncatedSVD, PCA
from sklearn.preprocessing import MinMaxScaler
from sklearn.compose import TransformedTargetRegressor
from gordo_components.model.anomaly.diff import DiffBasedAnomalyDetector

from gordo_components.model.models import KerasAutoEncoder
from gordo_components import serializer


logger = logging.getLogger(__name__)


class PipelineSerializationTestCase(unittest.TestCase):
    def test_pipeline_serialization(self):

        pipe = Pipeline(
            [
                ("pca1", PCA(n_components=10)),
                (
                    "fu",
                    FeatureUnion(
                        [
                            ("pca2", PCA(n_components=3)),
                            (
                                "pipe",
                                Pipeline(
                                    [
                                        ("minmax", MinMaxScaler()),
                                        ("truncsvd", TruncatedSVD(n_components=7)),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
            ]
        )

        X = np.random.random(size=100).reshape(10, 10)
        pipe.fit(X.copy(), X.copy())

        with TemporaryDirectory() as tmp:

            # Test dump
            metadata = {"key": "value"}
            serializer.dump(pipe, tmp, metadata=metadata)

            # Test load from the serialized pipeline above
            pipe_clone = serializer.load(tmp)
            metadata_clone = serializer.load_metadata(tmp)

            # Ensure the metadata was saved and loaded back
            self.assertEqual(metadata, metadata_clone)

            # Verify same state for both pipelines
            y_hat_pipe1 = pipe.predict(X.copy()).flatten()
            y_hat_pipe2 = pipe_clone.predict(X.copy()).flatten()
            self.assertTrue(np.allclose(y_hat_pipe1, y_hat_pipe2))

            # Now use dumps/loads
            serialized = serializer.dumps(pipe)
            pipe_clone = serializer.loads(serialized)

            # Verify same state for both pipelines
            y_hat_pipe1 = pipe.predict(X.copy()).flatten()
            y_hat_pipe2 = pipe_clone.predict(X.copy()).flatten()
            self.assertTrue(np.allclose(y_hat_pipe1, y_hat_pipe2))


@pytest.mark.parametrize(
    "model",
    [
        KerasAutoEncoder(kind="feedforward_hourglass"),
        DiffBasedAnomalyDetector(
            base_estimator=TransformedTargetRegressor(
                regressor=KerasAutoEncoder(kind="feedforward_symmetric"),
                transformer=MinMaxScaler(),
            )
        ),
        TransformedTargetRegressor(
            regressor=Pipeline(
                steps=[
                    ("stp1", MinMaxScaler()),
                    ("stp2", KerasAutoEncoder(kind="feedforward_symmetric")),
                ]
            )
        ),
    ],
)
def test_dump_load_models(model):

    X = np.random.random(size=100).reshape(10, 10)
    model.fit(X.copy(), X.copy())
    model_out = model.predict(X.copy())

    with TemporaryDirectory() as tmp:
        serializer.dump(model, tmp)

        model_clone = serializer.load(tmp)
        model_clone_out = model_clone.predict(X.copy())

        assert np.allclose(model_out.flatten(), model_clone_out.flatten())



import pytest

from gordo_components.dataset.sensor_tag import (
    normalize_sensor_tags,
    SensorTagNormalizationError,
)
from gordo_components.dataset.sensor_tag import SensorTag


TAG_NAME1 = "MyBeautifulTag1"
TAG_NAME2 = "MyBeautifulTag2"
asset_nonsense = "ImaginaryAsset"


@pytest.mark.parametrize(
    "good_input_tags,asset,expected_output_tags",
    [
        (
            [
                {"name": TAG_NAME1, "asset": asset_nonsense},
                {"name": TAG_NAME2, "asset": asset_nonsense},
            ],
            "ThisAssetCodeWillBeIgnored",
            [
                SensorTag(TAG_NAME1, asset_nonsense),
                SensorTag(TAG_NAME2, asset_nonsense),
            ],
        ),
        (
            ["TRC-123", "GRA-214", "ASGB-212"],
            "ThisWillBeTheAsset",
            [
                SensorTag("TRC-123", "ThisWillBeTheAsset"),
                SensorTag("GRA-214", "ThisWillBeTheAsset"),
                SensorTag("ASGB-212", "ThisWillBeTheAsset"),
            ],
        ),
        (
            ["TRC-123", "GRA-214", "ASGB-212"],
            None,  # Will deduce asset
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag("GRA-214", "1755-gra"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
        ),
        (
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag("GRA-214", "1755-gra"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
            "CouldWriteAssetHereButWeDontCare",
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag("GRA-214", "1755-gra"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
        ),
        (
            [[TAG_NAME1, asset_nonsense], [TAG_NAME2, asset_nonsense]],
            None,
            [
                SensorTag(TAG_NAME1, asset_nonsense),
                SensorTag(TAG_NAME2, asset_nonsense),
            ],
        ),
    ],
)
def test_normalize_sensor_tags_ok(good_input_tags, asset, expected_output_tags):
    tag_list_as_list_of_sensor_tag = normalize_sensor_tags(good_input_tags, asset)
    assert tag_list_as_list_of_sensor_tag == expected_output_tags


def test_normalize_sensor_tags_not_ok():
    with pytest.raises(SensorTagNormalizationError):
        tag_list_as_list_of_strings_nonsense = [TAG_NAME1, TAG_NAME2]
        normalize_sensor_tags(tag_list_as_list_of_strings_nonsense)



from gordo_components.client.utils import EndpointMetadata
from gordo_components.dataset.sensor_tag import SensorTag


def test_endpoint_metadata():
    data = {
        "endpoint-metadata": {
            "metadata": {
                "name": "test-project-target-a",
                "dataset": {
                    "resolution": "10T",
                    "tag_list": [],
                    "target_tag_list": [{"asset": "foo", "name": "bar"}],
                },
                "model": {"model-offset": 1},
            }
        },
        "healthy": True,
        "endpoint": "/gordo/v0/test-project/test-project-target-a",
    }

    epm = EndpointMetadata(data)
    assert epm.name == "test-project-target-a"
    assert epm.endpoint == "/gordo/v0/test-project/test-project-target-a"
    assert epm.tag_list == []
    assert epm.target_tag_list == [SensorTag(name="bar", asset="foo")]
    assert epm.resolution == "10T"
    assert epm.model_offset == 1
    assert epm.healthy is True
    assert epm.raw_metadata() == data



# -*- coding: utf-8 -*-

import unittest
import datetime

from gordo_components.workflow.config_elements.validators import (
    ValidUrlString,
    ValidDatetime,
    ValidTagList,
    ValidMetadata,
    ValidModel,
    ValidMachineRuntime,
    fix_resource_limits,
)


class DescriptorsTestCase(unittest.TestCase):
    def test_valid_model(self):
        """
        Verifies a given object is Union[str, dict]
        """

        class MyClass:
            value = ValidModel()

        myclass = MyClass()

        # All ok
        myclass.value = {
            "gordo_components.model.models.KerasAutoEncoder": {
                "kind": "feedforward_hourglass"
            }
        }
        myclass.value = "sklearn.ensemble.forest.RandomForestRegressor"

        # Not ok
        with self.assertRaises(ValueError):
            myclass.value = 1
        with self.assertRaises(ValueError):
            myclass.value = None

    def test_valid_metadata(self):
        """
        Verifies a given object is Optional[dict]
        """

        class MyClass:
            value = ValidMetadata()

        myclass = MyClass()

        # All of these are ok
        myclass.value = dict()
        myclass.value = {"key": "value"}
        myclass.value = None

        # This is not
        with self.assertRaises(ValueError):
            myclass.value = 1
        with self.assertRaises(ValueError):
            myclass.value = "string"

    def test_valid_datetime(self):
        """
        Verifies a given object is of datetime.datetime
        """

        class MyClass:
            value = ValidDatetime()

        myclass = MyClass()

        myclass.value = datetime.datetime.now(tz=datetime.timezone.utc)
        # Should raise an error
        with self.assertRaises(ValueError):
            myclass.value = datetime.datetime.now()

        # Should raise an error
        with self.assertRaises(ValueError):
            myclass.value = "not a datetime object"

    def test_valid_tag_list(self):
        """
        Verifies a given object is of a non-empty List[str]
        """

        class MyClass:
            value = ValidTagList()

        myclass = MyClass()

        # No problem
        myclass.value = ["string here", "string there"]

        # Problem.
        with self.assertRaises(ValueError):
            myclass.value = "not a list"
        with self.assertRaises(ValueError):
            myclass.value = []  # Error with empty list.

    def test_valid_url_string(self):
        """
        Verifies that a given string matches what we consider to be a valid
        url; in this case alphanumeric with dashes.
        """
        valid_names = [
            "valid-name-here",
            "validnamehere",
            "also-a-valid-name123",
            "equally-valid-name",
            "another-1-2-3",
        ]
        for valid_name in valid_names:
            self.assertTrue(
                ValidUrlString.valid_url_string(valid_name),
                msg=f"Expected '{valid_name}' to be valid, but it is not",
            )

        invalid_names = [
            "Not_a_valid_name",
            "C%tainly-not-v@lid",
            "also not valid with spaces",
            "couldn't-possibly-valid",
            "also,-this-is-not-valid",
            "(this-is-not-either)",
            "!nor-this",
            "cannot-have-UpperCase",
            "-cannot-begin-with-dashes",
            "cannot-end-with-dashes-",
        ]
        for invalid_name in invalid_names:
            self.assertFalse(
                ValidUrlString.valid_url_string(invalid_name),
                msg=f"Expected '{invalid_name}' to be invalid, but it is.",
            )

        # Test it's a working Descriptor
        class MyClass:
            value = ValidUrlString()

        myclass = MyClass()

        myclass.value = "this-should-not-cause-an-error"

        with self.assertRaises(ValueError):
            myclass.value = (
                "this-name-is-simply-too-long-that-is-the-only-problem-here-for-sure"
            )

        with self.assertRaises(ValueError):
            myclass.value = "but this should!"

    def test_valid_runtime(self):
        """
        Verifies that ValidRuntime enforces a server section with resources.requests and
        resources.limits
        """

        class MyClass:
            value = ValidMachineRuntime()

        myclass = MyClass()

        # Runtimes should be dicts
        myclass.value = {"influx": {"enable": True}}

        # Problem.
        with self.assertRaises(ValueError):
            myclass.value = 2

    def test_fix_resource_limits(self):
        # Bumps cpu limit
        resource_dict = {
            "requests": {"memory": 1, "cpu": 5},
            "limits": {"memory": 3, "cpu": 4},
        }
        res = fix_resource_limits(resource_dict)
        self.assertEqual(res["limits"]["cpu"], 5)

        # Works without requests
        resource_dict = {"limits": {"memory": 3, "cpu": 4}}
        res = fix_resource_limits(resource_dict)
        self.assertEqual(res["limits"]["cpu"], 4)
        # Works without limits
        resource_dict = {"requests": {"memory": 1, "cpu": 5}}
        res = fix_resource_limits(resource_dict)
        self.assertEqual(res["requests"]["cpu"], 5)

        # Does  nothing if limit>requests
        resource_dict = {
            "requests": {"memory": 1, "cpu": 5},
            "limits": {"memory": 3, "cpu": 6},
        }
        res = fix_resource_limits(resource_dict)
        self.assertEqual(res["limits"]["cpu"], 6)

    def test_fix_resource_limits_checks_ints(self):
        resource_dict = {
            "requests": {"memory": "1M", "cpu": 5},
            "limits": {"memory": 3, "cpu": 4},
        }
        with self.assertRaises(ValueError):
            fix_resource_limits(resource_dict)
